from gurobipy import *
from time import *
import time
from igraph import *
import sys
import argparse
import read_data_white as read
import irrelevant as irr

parser = argparse.ArgumentParser(description='ILP to find an optimal solution for Group Coverage Centrality.')
parser.add_argument('graph_file', type=str, help='graph file')
parser.add_argument('parameter', type=int, help='group size')
parser.add_argument('rrule', type=int,
                    help='usage of reduction rule to identify irrelevant vertices (1 for yes and each other value for no)')
parser.add_argument('time_limit', type=int, help='time limit in seconds for each instance')
parser.add_argument('outputfile', type=str,
                    help='name of the output file in which a new line is added with the result of this computation')
parser.add_argument('startsol', type=str,
                    help='A user may give the ILP an initial (greedy) solution. It has the form "[name1,name2,...,namek]. Otherwise, this will be ignored"')
args = parser.parse_args()

data_file = args.graph_file
k = args.parameter
all_covered = False

# load file and built graph
file = open(data_file.rstrip())
name = str(data_file).split('/')[-1]
graph = read.read_data(file)
graph.simplify()
print("Graph " + name + " has " + str(graph.vcount()) + " many vertices and " + str(graph.ecount()) + " many edges.")
file.close()
time_ok = True
open_neighborhoods = dict()
n = graph.vcount()

# read initial greedy solution
start_sol = False
init_ids = []
if args.startsol[0] == '[' and args.startsol[-1] == ']':
    start_sol = True
    init_sol = args.startsol
    # get rid of "[" and "]" and star and end
    init_sol = init_sol[1:-1]
    # get rif of "," as the seperator and save in list
    init_names = set(init_sol.split(","))
    # find the ids of the initial names and add them to init_ids
    for index in range(n):
        if graph.vs[index]["name"] in init_names:
            init_ids.append(index)

start_time = time_ns()
numb_irr = "not used"
time_irr = 0
irrr = [0 for i in range(n)]

# apply reduction rules to identify irrelevant vertices (if this option is activated)
if args.rrule == 1:
    graph, open_neighborhoods, numb_irr, time_irr, irrr, time_ok = irr.neighborhood_relation(graph, args.time_limit)
    if n - numb_irr <= k:
        all_covered = True
else:
    # computation of open neighborhoods because of excessive use in BFS
    for index in range(n):
        open_neighborhoods[index] = (set(graph.neighbors(index)))

# BFS-Tree for each vertex
# for each vertex u we have a list of lists where each list contains the vertices at some distance from u
bfs_distances = [[set()] for i in range(n)]
# also: distancematrix
distance_matrix = [[0] * n for i in range(n)]
# for each vertex u we have a list to store how many vertices have distance at most s to u
# this is later useful to determine all vertices on some shortest u-w path efficiently
distances_at_most = [[1] for i in range(n)]


# BFS
def bfs(visited, startv):
    visited[startv] = 0
    # only start vertex has distance 0 to itself
    bfs_distances[startv][0].add(startv)
    max_dist = 0
    queue = [startv]

    while queue:
        s = queue.pop(0)
        for neighbor in open_neighborhoods[s]:
            if visited[neighbor] == -1:
                new_dist = visited[s] + 1
                visited[neighbor] = new_dist
                if new_dist > max_dist:
                    max_dist = new_dist
                    bfs_distances[startv].append(set())
                if irrr[neighbor] == 0:
                    bfs_distances[startv][new_dist].add(neighbor)
                    queue.append(neighbor)
                distance_matrix[startv][neighbor] = new_dist


# output file
log_file = open(args.outputfile, 'a+')

# only continue with APSP computation if time limit is not yet reached
if time_ok:
    # if k is at least as large as the number of relevant vertices, we can immediately finish
    if all_covered == True:
        # compute solution
        solution_names = []
        for vertex in range(n):
            if irrr[vertex] == 0:
                solution_names.append(graph.vs[vertex]["name"])
        # compute number of covered vertex pairs
        numb_covered = n * (n - 1) / 2 - graph.ecount()
        tot_running_time = (time_ns() - start_time) / 1000000000
        out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(
            3) + str(numb_covered).rjust(10) + str(solution_names).replace(' ', '').rjust(k * 8) + str(
            tot_running_time).rjust(20) + "NA".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + str(
            time_irr).rjust(20) + str(numb_irr).rjust(10) + str(all_covered).rjust(6) + "\n"

    # interesting case: there are more than k relevant vertices
    else:
        # now start APSP computation
        # BFS starting from each vertex
        time_before_apsp = time_ns()
        for vertex in range(n):
            bfs([-1 for j in range(n)], vertex)
            # now compute the number of vertices at distance at most s from u for each possible s
            max_len = bfs_distances[vertex].__len__()
            for dist in range(1, max_len):
                distances_at_most[vertex].append(bfs_distances[vertex][dist].__len__())

        # update on time
        time_apsp = (time_ns() - time_before_apsp) / 1000000000
        print("Computed APSP in ", time.process_time(), "seconds.")

        # now, compute for each vertex u the set of vertex pairs covered by u
        # for each vertex u we save the vertex pairs which contain u on some shortest path
        time_before_vertices_on_shortest = time_ns()
        shortest_paths_covered = [[[] for j in range(n)] for i in range(n)]
        for u in range(n):
            for w in range(n):
                if w >= u:
                    break
                # only consider non-adjacent vertex pairs
                if distance_matrix[u][w] <= 1:
                    continue
                dist_u_w = distance_matrix[u][w]

                # for each i in range(1, dist_u_w) we compute the intersection of vertices of vertices at distance i to u and dist_u_w - i to w
                # the common neighbors are all neighbors on shortest paths
                for i in range(1, dist_u_w):
                    size_u = distances_at_most[u][i]
                    size_w = distances_at_most[w][dist_u_w - i]
                    smaller = u
                    bigger = w
                    index_small = i
                    index_big = dist_u_w - i
                    if size_w < size_u:
                        smaller = w
                        bigger = u
                        index_small = dist_u_w - i
                        index_big = i
                    common_neigs = bfs_distances[bigger][index_big].intersection(bfs_distances[smaller][index_small])
                    for vertex in common_neigs:
                        shortest_paths_covered[u][w].append(vertex)

        # update on time
        time_for_shortest_paths = (time_ns() - time_before_vertices_on_shortest) / 1000000000
        print("Computed all vertices on some shortest path for each non-adjacent vertex pair in ", time.process_time(),
              "seconds.")

        # Initialize ILP
        print('All prepared! Starting with Initializing the ILP..')
        print("Until now, it took " + str((time_ns() - start_time) / 1000000000) + "sec")
        opt_mod = Model(name="ILP for " + name)

        node_variables = []
        path_variables = []
        node_to_variable = dict()

        time_before_ilp_created = time_ns()
        for vertex in range(n):
            if irrr[vertex] == 1:
                continue
            new_variable = opt_mod.addVar(name=str(vertex), vtype=GRB.BINARY, lb=0, ub=1)
            node_variables.append(new_variable)
            node_to_variable[vertex] = new_variable
        opt_mod.update()

        # use initial greedy solution, if available
        if start_sol:
            for index in init_ids:
                # find corresponding gurobi variable
                if irrr[index] != 1:
                    val_x = opt_mod.getVarByName(str(index))
                    val_x.start = 1

        for v1 in range(n):
            for v2 in range(n):
                if v2 >= v1:
                    break
                # only constraints or non-adjacent vertex pairs
                if distance_matrix[v1][v2] == 1:
                    continue

                new_path_variable = opt_mod.addVar(name='path' + str(v1) + '/' + str(v2), vtype=GRB.BINARY, lb=0, ub=1)
                path_variables.append(new_path_variable)
                covering_node_variables = []
                for vertex in shortest_paths_covered[v1][v2]:
                    covering_node_variables.append(node_to_variable.get(vertex))

                opt_mod.addConstr(new_path_variable <= quicksum(covering_node_variables),
                                  name='ConstrP' + str(v1) + "/" + str(v2))

        opt_mod.addConstr(quicksum(node_variables) <= k, name='KConstr')

        opt_mod.setObjective(quicksum(path_variables), GRB.MAXIMIZE)
        time_to_create_ilp = (time_ns() - time_before_ilp_created) / 1000000000
        # only 1 thread + time limit
        opt_mod.Params.Threads = 1
        opt_mod.setParam('TimeLimit', max(1, args.time_limit - time_to_create_ilp))

        # now: solve the ILP
        print('ILP initialized. Now starting to solve it')

        time_before_ilp_solved = time_ns()
        opt_mod.optimize()
        time_to_solve_ilp = (time_ns() - time_before_ilp_solved) / 1000000000

        print('Done with solving the ILP')
        tot_running_time = (time_ns() - start_time) / 1000000000
        print("Total running time " + str(tot_running_time) + "sec")

        # stuff for output
        # data in output line:
        #
        # name
        # n
        # m
        # k
        # solution
        # total running time of the algo
        # time to solve the ILP
        # time to create the ILP
        # time for computation of all vertices on some shortest path for each non-adjacent vertex pair
        # time for APSP computation
        # time for computation of irrelevant vertices
        # number of irrelevant vertices
        # indicates whether all shortest paths are covered (True if yes and no otherwise)

        # check whether timelimit was reached
        if opt_mod.Status == 9:
            # time limit reached during solving the ILP
            out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(
                3) + "-1".rjust(10) + "[]".replace(' ', '').rjust(k * 8) + str(tot_running_time).rjust(20) + "time_limit_reached".rjust(20) + str(time_to_create_ilp).rjust(
                20) + str(
                time_for_shortest_paths).rjust(20) + str(time_apsp).rjust(20) + str(time_irr).rjust(20) + str(
                numb_irr).rjust(10) + str(all_covered).rjust(6) + "\n"
        else:
            # time limit was enough
            # determine solution
            solution = []
            for v in node_variables:
                if v.x == 1:
                    solution.append(graph.vs[int(v.varName)]["name"])

            out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(
                3) + str(opt_mod.objVal).rjust(15) + str(solution).replace(' ', '').rjust(k * 8) + str(
                tot_running_time).rjust(20) + str(time_to_solve_ilp).rjust(20) + str(time_to_create_ilp).rjust(
                20) + str(
                time_for_shortest_paths).rjust(20) + str(time_apsp).rjust(20) + str(time_irr).rjust(20) + str(
                numb_irr).rjust(10) + str(all_covered).rjust(6) + "\n"

            print('Objective Function Value: %f' % opt_mod.objVal)
            print('Nodes in the Set (IDs):')
            for v in node_variables:
                if v.x == 1:
                    print(v.varName)

# time limit reached in computation of N(u) subseteq N[w]
else:
    out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(3) + "not solved".rjust(10) + "".rjust(50) + "time limit reached".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + "NA".rjust(20) + str(numb_irr).rjust(10) + str(all_covered).rjust(6) + "\n"

log_file.write(out_line)
log_file.close()
