import sys
sys.setrecursionlimit(10000000)
from time import *
import time
import random as random
from igraph import *
import argparse
import read_data_white as read

parser = argparse.ArgumentParser(description='ILP to find an optimal solution for Group Coverage Centrality.')
parser.add_argument('graph_file', type=str, help='graph file')
parser.add_argument('parameter', type=int, help='group size')
parser.add_argument('variant', type=int, help='Specifies which greedy algorithm is used')
parser.add_argument('time_limit', type=int, help='time limit in seconds for each instance')
parser.add_argument('outputfile', type=str, help='name of the output file in which a new line is added with the result of this computation')
parser.add_argument('quality', type=int, help='If 1, the number of covered vertex pairs is computed, otherwise only the solution is outputted')
args = parser.parse_args()

data_file = args.graph_file
k = args.parameter
time_limit = args.time_limit

# load file and built graph
file = open(data_file.rstrip())
name = str(data_file).split('/')[-1]
graph = read.read_data(file)
#print(max(graph.degree()))
graph.simplify()
n = graph.vcount()
print("Graph " + name + " has " + str(n) + " many vertices and " + str(graph.ecount()) + " many edges.")
file.close()

# call of the chosen heuristic
my_var = args.variant

# time
my_time = 0


# pick vertices with high degree
# always update the set of neighbors after a vertex is chosen to be in the solution
def greedyDegree(graph, k):
    uncovered_neighbors = dict()
    # compute T[v] sets for each vertex
    for index in range(n):
        uncovered_neighbors[index] = set(graph.neighbors(index)).union({index})

    solution_nodes = set()
    sol_size = 0
    # find vertex w which maximizes |T[w]| if there is any x with |T[x]| >= 1
    while sol_size < k:
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        candidate = None
        maximum = 0
        for index in range(n):
            if uncovered_neighbors[index] is not None and len(uncovered_neighbors[index]) > maximum:
                maximum = len(uncovered_neighbors[index])
                candidate = index
        if candidate is None:
            break
        # update solution with best candidate
        sol_size += 1
        solution_nodes.add(candidate)
        covered_nodes = set(graph.neighbors(candidate)).union({candidate})
        # update T[x] values
        for index in range(n):
            uncovered_neighbors[index] = uncovered_neighbors[index] - covered_nodes

    # if solution is not big enough we are in the case that |T[x]| = 0 for each x
    # now, greedily add the vertices with the highest degree to the solution
    if sol_size < k:
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        degrees = graph.degree()
        indices = range(n)
        degree_ordering = sorted(list(zip(degrees, indices)), reverse=True)
        while sol_size < k:
            candidate = degree_ordering.pop(0)
            if candidate[1] not in solution_nodes:
                sol_size += 1
                solution_nodes.add(candidate[1])

    return solution_nodes, False


# greedily pick vertex which has the highest number of non-neighbors in his neighborhood
# update these values after each step
def greedyNonNeighhbors(graph, k):
    # compute open neighbors of each vertex
    open_neighborhoods = dict()
    for index in range(n):
        open_neighborhoods[index] = (set(graph.neighbors(index)))
    # number and pairs of neighbors which are not connected
    not_connected_neighbors = dict()
    not_connected_neighbor_pairs = dict()
    contained = set()
    # compute these values for each vertex
    for index in range(n):
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        count = 0
        not_connected_neighbor_pairs[index] = set()
        for neighbor1 in open_neighborhoods[index]:
            for neighbor2 in open_neighborhoods[index]:
                if neighbor1 < neighbor2 and not neighbor1 in open_neighborhoods[neighbor2]:
                    count += 1
                    if count > 0:
                        not_connected_neighbor_pairs[index].add((min(neighbor1, neighbor2), max(neighbor1, neighbor2)))
        if count > 0:
            not_connected_neighbors[index] = count
            contained.add(index)

    # now compute greedy solution
    solution_nodes = set()
    sol_size = 0
    while sol_size < k:
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        # find vertex with maximal number of non-neighbors in his neighborhood
        maximum = 0
        candidate = None
        for index in contained:
            numb_non_neigs = not_connected_neighbors[index]
            if maximum < numb_non_neigs:
                maximum = numb_non_neigs
                candidate = index

        # only go further if there is at least one vertex with at least one non-edge in his neighborhood
        if candidate is None:
            break
        # add candidate to solution
        solution_nodes.add(candidate)
        sol_size += 1
        covered_non_neigs = not_connected_neighbor_pairs[candidate]
        # update the number and pairs of non-neighbors
        for index in range(n):
            not_connected_neighbor_pairs[index] = not_connected_neighbor_pairs[index] - covered_non_neigs
            not_connected_neighbors[index] = len(not_connected_neighbor_pairs[index])

    # if solution is not big enough we are in the case that all vertex pairs of distance 2 are covered
    # now, greedily add the vertices with the highest degree to the solution
    if sol_size < k:
        degrees = graph.degree()
        indices = range(n)
        degree_ordering = sorted(list(zip(degrees, indices)), reverse=True)
        while sol_size < k:
            candidate = degree_ordering.pop(0)
            if candidate[1] not in solution_nodes:
                sol_size += 1
                solution_nodes.add(candidate[1])

    return solution_nodes, False

# calculate binom(z,2)
def binom2(z):
    return z*(z-1)/2


# greedily pick a vertex which has many neighbors and two-neighbors
# precise score of vertex v is:            we use: b(a,c) = \binom(a)(c)
# b(N(v),2) + b(N_2(v),2) - sum_(w\in N(v)) deg(w) - \sum_(w\in N_2(v)) deg(w)
# we update these values
def greedyTwoNeighhbors(graph, k):
    # compute open neighbors of each vertex
    open_neighborhoods = dict()
    neighbor_size = dict()
    for index in range(n):
        neigs = set(graph.neighbors(index))
        open_neighborhoods[index] = (neigs)
        neighbor_size[index] = len(neigs)

    twoneighbor_size = dict()
    # score of each vertex
    scores = dict()
    for index in range(n):
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        my_score = binom2(neighbor_size[index])
        second_neigs = set()
        this_neigs = open_neighborhoods[index]
        for neig in this_neigs:
            #my_score -= binom2(neighbor_size[neig])
            my_score -= neighbor_size[neig]
            second_neigs.update(open_neighborhoods[neig])
        two_neig_size = len(second_neigs)
        my_score += binom2(two_neig_size)
        twoneighbor_size[index] = two_neig_size
        for two_neig in second_neigs:
            my_score -= neighbor_size[two_neig]
        scores[index] = my_score

    # determine solution
    solution_nodes = set()
    # k times: pick vertex with the highest score
    # after each pick, update the scores of the remaining vertices
    for index in range(k):
        candidate = None
        best_score = -sys.maxsize - 1
        for vertex in range(n):
            my_score = scores[vertex]
            if vertex not in solution_nodes and my_score > best_score:
                candidate = vertex
                best_score = my_score
        solution_nodes.add(candidate)


        # compute sum of neighbors of candidate
        sum_neigs_cand = 0
        two_neigs_cand = set()
        for neig in open_neighborhoods[candidate]:
            sum_neigs_cand += neighbor_size[neig]
            two_neigs_cand.update(open_neighborhoods[neig])

        # update scores of neighbors
        for neig in open_neighborhoods[candidate]:
            new_two_size = twoneighbor_size[neig] - neighbor_size[candidate]
            scores[neig] += neighbor_size[candidate] - sum_neigs_cand - neighbor_size[neig] - binom2(twoneighbor_size[neig]) + binom2(new_two_size)
            twoneighbor_size[neig] = new_two_size

        # update scores of second neighbors
        for two_neig in two_neigs_cand:
            scores[two_neig] += neighbor_size[candidate] - twoneighbor_size[two_neig]

    return solution_nodes, False

# compute all cut vertices of G
# let v be a cut vertex with cut sets of size A, B, C and D
# value(v) = AB + AC + AD + BC + BD + CD
# pick k vertices maximizing this value, without updates
# if there are not enough cut-vertices, pick vertices with the highest degree
def greedyCut(graph, k):
    # compute open neighbors of each vertex
    open_neighborhoods = dict()
    for index in range(n):
        open_neighborhoods[index] = (set(graph.neighbors(index)))

    # The function to do DFS traversal.
    # initially, no vertex is yet discovered
    # also initialize empty parent array and empty list of articulation points and there value list
    visited = [False for i in range(n)]
    discover_times = [sys.maxsize for i in range(n)]
    lowest = [sys.maxsize for i in range(n)]
    parents = [-1 for i in range(n)]
    articulation_points = set()

    # u: is current vertex in the DFS call
    # lowest: earliest vertex visited by the algo (that is according to the discover_time) which can be reached from the subtree rooted at the current vertex
    # discover_times: time when a vertex got discovered
    # parents: indices of the parent in the DFS tree
    # articulation_points: list of articulation points
    # articulation_points_vals: values of the articulation points
    def DFS_call(u, lowest, discover_times, visited, parents, articulation_points):
        # Count of children in current node
        numb_children = 0
        # current vertex is now visited
        visited[u] = True

        # set discovery time and set initial lowest value
        global my_time
        discover_times[u] = my_time
        lowest[u] = my_time
        my_time += 1

        # DFS for each neighbor
        for v in open_neighborhoods[u]:
            # only go further at v, if v is not yet visited
            if not visited[v]:
                # set information about v
                parents[v] = u
                numb_children += 1
                # call DFS iteratively
                DFS_call(v, lowest, discover_times, visited, parents, articulation_points)

                # check if the subtree rooted with v has a connection to one of the ancestors of u
                lowest[u] = min(lowest[u], lowest[v])

                # if u is an articulation point, one of the two cases occur:
                # a) u is root of DFS tree and has two or more children
                if parents[u] == -1 and numb_children > 1:
                    articulation_points.add(u)

                # b) if u is not root and low value of one of its child is more than discovery value of u
                if parents[u] != -1 and lowest[v] >= discover_times[u]:
                    articulation_points.add(u)

            # finally, update low value of u for parent function calls
            elif v != parents[u]:
                lowest[u] = min(lowest[u], discover_times[v])

    # now call DFS to compute these values
    for index in range(n):
        if not visited[index]:
            # check time limit:
            runtime = (time_ns() - start_time) / 1000000000
            if runtime > time_limit:
                return set(), True
            DFS_call(index, lowest, discover_times, visited, parents, articulation_points)

    # now compute cut sizes of each articulation point to compute the respective value

    # BFS to determine the sizes of the partite sets
    # returns number of vertices in this component
    def bfs3(visited, startv):
        visited[startv] = 0
        queue = [startv]
        compo_size = 0

        while queue:
            s = queue.pop(0)
            compo_size += 1
            for neighbor in open_neighborhoods[s]:
                if visited[neighbor] == -1:
                    visited[neighbor] = 0
                    queue.append(neighbor)

        return compo_size

    # save articulation points and their respective partite sets
    art_list = []
    art_vals = []
    for art_point in articulation_points:
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), True
        # perform BFS from each neighbor of the articulation point to determine the cut sizes
        art_list.append(art_point)
        art_compos = []
        visited = [-1 for i in range(n)]
        visited[art_point] = 0
        for neig in open_neighborhoods[art_point]:
            if visited[neig] == -1:
                compo_size = bfs3(visited, neig)
                art_compos.append(compo_size)

        # now compute the values of each articulation point
        mysum = 0
        myres = 0
        cur_size = len(art_compos) - 1

        while cur_size >= 0:
            myres += art_compos[cur_size] * mysum
            mysum += art_compos[cur_size]
            cur_size -= 1
        art_vals.append(myres)

    print("done with precompution")
    # now greedily pick the vertices according to the cut size
    solution = set()
    # Case 1: there are more than k cut vertices
    if len(art_list) > k:
        art_point_ordering = sorted(list(zip(art_vals, art_list)), reverse=True)
        print("many cuts")
        for index in range(k):
            best = art_point_ordering.pop(0)
            solution.add(best[1])
    # otherwise there are at most k cut vertices
    # the greedy algorithm chooses all of them and for the rest it picks vertices with maximal degree
    else:
        for index in art_list:
            solution.add(index)

        # for the rest: use vertices with high degree
        print("less cuts")
        sol_size = len(solution)
        degrees = graph.degree()
        indices = range(n)
        degree_ordering = sorted(list(zip(degrees, indices)), reverse=True)
        while sol_size < k:
            candidate = degree_ordering.pop(0)
            if candidate[1] not in solution:
                sol_size += 1
                solution.add(candidate[1])
    return solution, False


# build the corresponding set cover instance
# elements are the non-adjacent vertex pairs and sets are the vertices
# there is an edge between element e and set F=(s,t) if vertex e is on a shortest s-t-path
# now: greedy set cover algorithm
# always pick set which covers the maximal number of elements and update after each setp
def greedySetCover(graph, k):
    # compute corresponding set cover instance and then greedily pick most profitable sets

    open_neighborhoods = dict()
    for index in range(n):
        open_neighborhoods[index] = (set(graph.neighbors(index)))

    # BFS-Tree for each vertex
    # for each vertex u we have a list of lists where each list contains the vertices at some distance from u
    bfs_distances = [[set()] for i in range(n)]
    # also: distancematrix
    distance_matrix = [[0] * n for i in range(n)]
    # for each vertex u we have a list to store how many vertices have distance at most s to u
    # this is later useful to determine all vertices on some shortest u-w path efficiently
    distances_at_most = [[1] for i in range(n)]

    # BFS with storage of bfs-distances
    # necessary for greedy set cover approach
    def bfs2(visited, startv):
        visited[startv] = 0
        # only start vertex has distance 0 to itself
        bfs_distances[startv][0].add(startv)
        max_dist = 0
        queue = [startv]

        while queue:
            s = queue.pop(0)
            for neighbor in open_neighborhoods[s]:
                if visited[neighbor] == -1:
                    new_dist = visited[s] + 1
                    visited[neighbor] = new_dist
                    if new_dist > max_dist:
                        max_dist = new_dist
                        bfs_distances[startv].append(set())
                    bfs_distances[startv][new_dist].add(neighbor)
                    queue.append(neighbor)
                    distance_matrix[startv][neighbor] = new_dist

    for vertex in range(n):
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), 0, True
        bfs2([-1 for j in range(n)], vertex)
        # now compute the number of vertices at distance at most s from u for each possible s
        max_len = bfs_distances[vertex].__len__()
        for dist in range(1, max_len):
            # distances_at_most[vertex].append(distances_at_most[vertex][-1] + bfs_distances[vertex][dist].__len__())
            distances_at_most[vertex].append(bfs_distances[vertex][dist].__len__())

    # each vertex has a list of pairs which are covered by him
    vertices_covered = [set() for i in range(n)]
    for u in range(n):
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), 0, True
        for w in range(n):
            if w >= u:
                break
            # only consider non-adjacent vertex pairs
            if distance_matrix[u][w] <= 1:
                continue
            dist_u_w = distance_matrix[u][w]

            # for each i in range(1, dist_u_w) we compute the intersection of vertices of vertices at distance i to u and dist_u_w - i to w
            # the common neighbors are all neighbors on shortest paths
            for i in range(1, dist_u_w):
                size_u = distances_at_most[u][i]
                size_w = distances_at_most[w][dist_u_w - i]
                smaller = u
                bigger = w
                index_small = i
                index_big = dist_u_w - i
                if size_w < size_u:
                    smaller = w
                    bigger = u
                    index_small = dist_u_w - i
                    index_big = i
                common_neigs = bfs_distances[bigger][index_big].intersection(bfs_distances[smaller][index_small])
                for vertex in common_neigs:
                    vertices_covered[vertex].add((w, u))

    # now greedy procedure
    # always chose a vertex with the highest coverage
    sol_size = 0
    coverage_so_far = 0
    solution = set()
    while sol_size < k:
        # check time limit:
        runtime = (time_ns() - start_time) / 1000000000
        if runtime > time_limit:
            return set(), 0, True
        candidate = None
        maximum = 0
        for index in range(n):
            if vertices_covered[index] is not None and len(vertices_covered[index]) > maximum:
                maximum = len(vertices_covered[index])
                candidate = index
        if candidate is None:
            break
        # adapt solution:
        sol_size += 1
        solution.add(candidate)
        coverage_so_far += len(vertices_covered[candidate])
        set_covered = vertices_covered[candidate]

        # update coverage of the vertices
        for index in range(n):
            vertices_covered[index] = vertices_covered[index] - set_covered
    return solution, coverage_so_far, False


# BFS
def bfs(visited, startv):
    visited[startv] = 0
    # only start vertex has distance 0 to itself
    max_dist = 0
    queue = [startv]

    while queue:
        s = queue.pop(0)
        for neighbor in open_neighborhoods[s]:
            if visited[neighbor] == -1:
                new_dist = visited[s] + 1
                visited[neighbor] = new_dist
                if new_dist > max_dist:
                    max_dist = new_dist
                queue.append(neighbor)
                distance_matrix[startv][neighbor] = new_dist


# determine number of covered vertex pairs
def numb_pairs_covered(solution):
    for vertex in range(n):
        bfs([-1 for j in range(n)], vertex)

    # now, compute for each vertex u the set of vertex pairs covered by u
    # for each vertex u we save the vertex pairs which contain u on some shortest path
    print("teeeeest " + str(my_var) + " " + str(solution) + " "  + name + " " + str(k))
    numb_covered = 0
    for u in range(n):
        for w in range(n):
            if w >= u:
                break
            # only consider non-adjacent vertex pairs
            if distance_matrix[u][w] <= 1:
                continue
            dist_u_w = distance_matrix[u][w]

            found = False
            for vertex in solution:
                if dist_u_w == distance_matrix[u][vertex] + distance_matrix[w][vertex]:
                    found = True
                    break
            if found:
                numb_covered += 1
    return numb_covered


mysol = None
pairs_covered = None
start_time = time_ns()
time_limit_reached = False
if my_var == 1:
    mysol, time_limit_reached = greedyDegree(graph, k)
elif my_var == 2:
    mysol, pairs_covered,time_limit_reached = greedySetCover(graph, k)
elif my_var == 3:
    mysol, time_limit_reached = greedyNonNeighhbors(graph, k)
elif my_var == 4:
    mysol, time_limit_reached = greedyTwoNeighhbors(graph, k)
else:
    mysol, time_limit_reached = greedyCut(graph, k)

print("computed " + name + " " + str(k) + " " + str(my_var) + " " + str(mysol))
print(graph.vcount(), graph.ecount())
# time to find that greedy solution
runtime = (time_ns() - start_time) / 1000000000

# write to output file
log_file = open(args.outputfile, 'a+')

# write to output that timelimit was not sufficient
time_string = "time_limit_of_"+str(time_limit)+"_seconds_reached"
if time_limit_reached is True:
    out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(3) + str(
        my_var).rjust(3) + "not_computed".rjust(14) + "not_computed".rjust(14) + time_string.rjust(40)+ "\n"
else:
    # compute number of covered vertex pairs, if option is activated
    if args.quality == 1 and pairs_covered is None:
        # BFS-Tree for each vertex
        # distancematrix
        distance_matrix = a = [[0] * n for i in range(n)]
        open_neighborhoods = dict()
        for index in range(n):
            open_neighborhoods[index] = (set(graph.neighbors(index)))
        pairs_covered = numb_pairs_covered(mysol)



    # transform internal IDs of solution into names of the input graph
    solution_names = []
    for index in mysol:
        solution_names.append(graph.vs[index]["name"])

    if pairs_covered != None:
        out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(3) + str(
            my_var).rjust(3) + str(pairs_covered).rjust(14) + str(solution_names).replace(' ', '').rjust(k * 8) + str(
            runtime).rjust(20) + "\n"
    else:
        out_line = name.rjust(40) + str(graph.vcount()).rjust(9) + str(graph.ecount()).rjust(11) + str(k).rjust(3) + str(
            my_var).rjust(3) + "not_computed".rjust(14) + str(solution_names).replace(' ', '').rjust(k * 8) + str(
            runtime).rjust(20) + "\n"

log_file.write(out_line)
log_file.close()
